from enum import IntEnum
from time import sleep

from . import logger
from .itla12 import ITLA12
from .itla_errors import (
    RVEError,
)


class Mode(IntEnum):
    normal = 0
    nodither = 1
    whisper = 2


class PPLaser(ITLA12):
    """
    The pure photonics laser class implements specific features or handles particular
    quirks of the pure photonics laser. The pure photonics laser has additional
    registers for a couple of applications.

    * `set_frequency` and `set_fcf` had to be overridden
    * `get_frequency` is *not* overridden because these registers can still be read without issue
    """

    def __init__(self, serial_port, baudrate=9600, sleep_time=0.1):
        """sets up the additional frequency max and min variables which will be set upon
        connecting to the laser. We have found that Pure Photonics lasers do not
        return the RVEError when setting the frequency out of spec.

        This sets up the additional register yaml file for pure photonics specific functions.
        """
        register_files = ["registers_pp.yaml"]

        self._frequency_max = None
        self._frequency_min = None

        super().__init__(
            serial_port, baudrate, register_files=register_files, sleep_time=sleep_time
        )

    def connect(self):
        """Overriden connect function with query for max and min frequency"""
        super().connect()
        self._frequency_max = self.get_frequency_max()
        self._frequency_min = self.get_frequency_min()

    def set_fcf(self, freq):
        """
        This sets the first channel frequency.
        It does not reset the channel so this frequency will only be equal
        to the output frequency if channel=1.

        Because the `set_frequency` function calls this function this will also
        correct the issue in setting the frequency generally.

        verifies that fcf is set within the appropriate laser frequency range
        and raises RVE error if not
        """
        if freq < self._frequency_min or freq > self._frequency_max:
            raise RVEError(
                "The desired frequency is outside of the range for this laser."
            )

        super().set_fcf(freq)

    def get_mode(self):
        """get which low noise mode"""

        response = self._mode()
        response = Mode(int.from_bytes(response, "big"))

        return response

    def set_mode(self, mode):
        self._mode(mode)

    def normalmode(self):
        """set mode to standard dither mode"""
        self._mode(Mode.normal)

    def nodithermode(self):
        """Set mode to nodither mode

        It is unclear whether this just also activates whisper mode.
        The feature guide says "a value of 1 defaults to 2".
        """
        self._mode(Mode.nodither)

    def whispermode(self):
        """Enables whisper mode where all control loops
        are disabled resulting in a lower noise mode."""
        self._mode(Mode.whisper)

    def get_cleansweep_amplitude(self):
        """get the amplitude of the clean sweep.

        The clean sweep will ramp to (+0.5 to -0.5) * amplitude.
        (Basically it will be centered at the current frequency.)
        """
        response = self._csrange()
        cs_amplitude = int.from_bytes(response, "big")

        return cs_amplitude

    def set_cleansweep_amplitude(self, range_GHz):
        """Sets the amplitude of the clean sweep.

        The clean sweep will ramp to (+0.5 to -0.5) * amplitude.
        (Basically it will be centered at the current frequency.)
        """
        self._csrange(range_GHz)

    def cleansweep_enable(self):
        """Enables clean sweep mode.
        The clean sweep will ramp to (+0.5 to -0.5) * amplitude.
        (Basically it will be centered at the current frequency.)
        """
        response = self.get_mode()
        if response == Mode.normal:
            raise Exception(
                "Laser currently in normal mode.",
                "please set to nodither or whisper mode",
            )

        self._csstart(1)

    def cleansweep_disable(self):
        """Turn's off clean sweep mode."""

        self._csstart(0)

    def cleansweep_setup(self, freq_low, freq_high):
        """
        A wrapper function that takes the start and end frequencies
        you would like to sweep over. It computes and sets the clean sweep
        amplitude and the midpoint.

        The clean sweep operation will begin in the center of the high and low.

        """

        if freq_low < self._frequency_min or freq_high > self._frequency_max:
            raise ValueError(
                "The range you would like to sweep over",
                "is outside of the bounds for this laser",
            )

        # sets the clean sweep amplitude rounded to the nearest MHz.
        cs_ampl = round(abs(freq_high - freq_low) * 1e3)
        if cs_ampl > 50:
            raise Exception(
                "This is an error specific to SpectrumLab."
                "Our laser has a 50GHz maximum cleansweep range."
                "You should change the limit in source for your laser."
                "If you know a way to query the laser for this value"
                "then implementing that would be better."
            )

        self.set_cleansweep_amplitude(cs_ampl)

        cs_center = round((freq_low + freq_high) * 1e4 / 2) * 1e-4
        self.set_frequency(cs_center)

    def get_cleansweep_rate(self):
        """Gets the clean sweep rate. not sure about units."""
        response = self._csrate()
        cs_rate = int.from_bytes(response, "big")
        return cs_rate

    def set_cleansweep_rate(self, rate_MHz):
        """sets the cleansweep rate in MHz/sec
        idk if there is a way to check if its within some bounds.
        There is conflicting info in the pure photonics docs."""
        raise Warning("Dont use this. Need to confirm the units for the sweep rate.")
        self._csrate(rate_MHz)

    def get_cleansweep_offset(self):
        raise Warning("This is not implemented yet.")

    def set_cleansweep_offset(self, offset):
        raise Warning("This is not implemented yet.")

    def cleanjump(self):
        """Tells the laser to perform the clean jump.
        The clean jump must be precalibrated.
        You should have the calibration points in your lab notebook
        or written down somewhere.
        If not you should perform the clean jump calibration."""
        mode = self.get_mode()
        if mode == Mode.normal:
            raise Exception(
                "Laser in normal mode. You must be in nodither"
                "or whisper mode to use cleanjump."
            )
        self._cjstart(1)

    def set_cleanjump(self, freq_jump):
        """
        Set the size of the jump in THz to the fourth decimal place.
        XXX.XXXX
        """
        freq_jump = str(round(freq_jump * 1e4))
        self._cjthz(int(freq_jump[0:3]))
        self._cjghz(int(freq_jump[3:]))

    def cleanjump_calibration(self, fcf, grid, pwr, n_jumppoints, confirmation=None):
        """
        This is a grid based calibration

        **fcf**: the first channel frequency you would like to
        begin calibration at. (THz)

        **grid**: the grid spacing (GHz)

        **pwr**: The power calibration will be at. (dBm)

        **n_jumppoints**: The number of grid points to calibrate.

        **confirmation**: If you would like to go ahead and confirm yes
        and skip the user input you can set confirmation to 'y'
        """

        # calibration begins by writing the number of channels to 0xD2

        calibration_points = [fcf + grid * n * 1e-3 for n in range(n_jumppoints)]

        logger.info(
            f"You are calibrating {n_jumppoints} clean jump setpoints",
            f"at frequencies:\n{calibration_points}(THz).",
            f"With power = {pwr}(dBm).",
        )
        logger.info(
            "Write these down in your lab notebook because",
            "you cannot query the device for these values later.",
        )

        if confirmation is None:
            confirmation = input("Is the information above correct? (y/n): ")

        if confirmation.lower() == "y":
            # set up the first channel frequency, grid spacing, and power
            self.set_frequency(fcf)
            self.set_grid(grid)
            self.set_power(pwr)
            self._cleanjump_calibration(n_jumppoints)

        elif confirmation.lower() == "n":
            logger.info(
                f"You answered {confirmation!r}.", "Calibration will not be performed."
            )

        else:
            raise Exception("input not recognized")

    def _cleanjump_calibration(self, n_jumppoints):
        # does this return a response with bit 15 active?
        self._cjcalibration(n_jumppoints)

        is_calibrating = True

        while is_calibrating:
            response = self._cjcalibration()

            response_bits = [bit for bit in f"{response:08}"].reverse()
            is_calibrating = bool(response_bits[15])
            sleep(1)
